@@ -50,7 +50,13 @@ if typing.TYPE_CHECKING:  # pragma: no cover
     import wheel.wheelfile  # noqa: F401
 
 
-__version__ = '0.8.1'
+if sys.version_info >= (3, 8):
+    from functools import cached_property
+else:
+    cached_property = lambda x: property(functools.lru_cache(maxsize=None)(x))  # noqa: E731
+
+
+__version__ = '0.9.0'
 
 
 class _depstr:
@@ -134,20 +140,38 @@ class _WheelBuilder():
     _SCHEME_MAP: ClassVar[Dict[str, Tuple[str, ...]]] = {
         'scripts': ('{bindir}',),
         'purelib': ('{py_purelib}',),
-        # {moduledir_shared} should be listed in platlib here too, but currently
-        # there is a Meson bug preventing us from using, so we will have to let
-        # that fallback on heuristics.
-        # see https://github.com/mesonbuild/meson/pull/9474
-        'platlib': ('{py_platlib}',),
+        'platlib': ('{py_platlib}', '{moduledir_shared}'),
         'headers': ('{includedir}',),
         'data': ('{datadir}',),
         # our custom location
         'mesonpy-libs': ('{libdir}', '{libdir_shared}')
     }
 
-    def __init__(self, project: Project) -> None:
+    def __init__(
+        self,
+        project: Project,
+        source_dir: pathlib.Path,
+        install_dir: pathlib.Path,
+        build_dir: pathlib.Path,
+        sources: Dict[str, Dict[str, Any]],
+        copy_files: Dict[str, str],
+    ) -> None:
         self._project = project
-        self._libs_build_dir = project._build_dir / 'mesonpy-wheel-libs'
+        self._source_dir = source_dir
+        self._install_dir = install_dir
+        self._build_dir = build_dir
+        self._sources = sources
+        self._copy_files = copy_files
+
+        self._libs_build_dir = self._build_dir / 'mesonpy-wheel-libs'
+
+    @cached_property
+    def _wheel_files(self) -> DefaultDict[str, List[Tuple[pathlib.Path, str]]]:
+        return self._map_to_wheel(self._sources, self._copy_files)
+
+    @property
+    def _has_internal_libs(self) -> bool:
+        return bool(self._wheel_files['mesonpy-libs'])
 
     @property
     def basename(self) -> str:
@@ -162,9 +186,9 @@ class _WheelBuilder():
         """Wheel name, this includes the basename and tags."""
         return '{basename}-{python_tag}-{abi_tag}-{platform_tag}'.format(
             basename=self.basename,
-            python_tag=self._project.python_tag,
-            abi_tag=self._project.abi_tag,
-            platform_tag=self._project.platform_tag,
+            python_tag=self.python_tag,
+            abi_tag=self.abi_tag,
+            platform_tag=self.platform_tag,
         )
 
     @property
@@ -175,6 +199,20 @@ class _WheelBuilder():
     def data_dir(self) -> str:
         return f'{self.basename}.data'
 
+    @cached_property
+    def is_pure(self) -> bool:
+        """Is the wheel "pure" (architecture independent)?"""
+        # XXX: I imagine some users might want to force the package to be
+        # non-pure, but I think it's better that we evaluate use-cases as they
+        # arise and make sure allowing the user to override this is indeed the
+        # best option for the use-case.
+        if self._wheel_files['platlib']:
+            return False
+        for _, file in self._wheel_files['scripts']:
+            if self._is_native(file):
+                return False
+        return True
+
     @property
     def wheel(self) -> bytes:  # noqa: F811
         """Return WHEEL file for dist-info."""
@@ -184,8 +222,8 @@ class _WheelBuilder():
             Root-Is-Purelib: {is_purelib}
             Tag: {tags}
         ''').strip().format(
-            is_purelib='true' if self._project.is_pure else 'false',
-            tags=f'{self._project.python_tag}-{self._project.abi_tag}-{self._project.platform_tag}',
+            is_purelib='true' if self.is_pure else 'false',
+            tags=f'{self.python_tag}-{self.abi_tag}-{self.platform_tag}',
         ).encode()
 
     @property
@@ -201,10 +239,184 @@ class _WheelBuilder():
         except ModuleNotFoundError:
             return False
 
-    def _is_elf(self, file: Union[str, pathlib.Path]) -> bool:
-        """Check if file is an ELF file."""
+    @property
+    def python_tag(self) -> str:
+        selected_tag = self._select_abi_tag()
+        if selected_tag and selected_tag.python:
+            return selected_tag.python
+        return 'py3'
+
+    @property
+    def abi_tag(self) -> str:
+        selected_tag = self._select_abi_tag()
+        if selected_tag:
+            return selected_tag.abi
+        return 'none'
+
+    @cached_property
+    def platform_tag(self) -> str:
+        if self.is_pure:
+            return 'any'
+        # XXX: Choose the sysconfig platform here and let something like auditwheel
+        #      fix it later if there are system dependencies (eg. replace it with a manylinux tag)
+        platform_ = sysconfig.get_platform()
+        parts = platform_.split('-')
+        if parts[0] == 'macosx':
+            target = os.environ.get('MACOSX_DEPLOYMENT_TARGET')
+            if target:
+                print(
+                    '{yellow}MACOSX_DEPLOYMENT_TARGET is set so we are setting the '
+                    'platform tag to {target}{reset}'.format(target=target, **_STYLES)
+                )
+                parts[1] = target
+            else:
+                # If no target macOS version is specified fallback to
+                # platform.mac_ver() instead of sysconfig.get_platform() as the
+                # latter specifies the target macOS version Python was built
+                # against.
+                parts[1] = platform.mac_ver()[0]
+
+            if parts[1] in ('11', '12'):
+                # Workaround for bug where pypa/packaging does not consider macOS
+                # tags without minor versions valid. Some Python flavors (Homebrew
+                # for example) on macOS started to do this in version 11, and
+                # pypa/packaging should handle things correctly from version 13 and
+                # forward, so we will add a 0 minor version to MacOS 11 and 12.
+                # https://github.com/FFY00/meson-python/issues/91
+                # https://github.com/pypa/packaging/issues/578
+                parts[1] += '.0'
+
+            platform_ = '-'.join(parts)
+        elif parts[0] == 'linux' and parts[1] == 'x86_64' and sys.maxsize == 0x7fffffff:
+            # 32-bit Python running on an x86_64 host
+            # https://github.com/FFY00/meson-python/issues/123
+            parts[1] = 'i686'
+            platform_ = '-'.join(parts)
+        return platform_.replace('-', '_').replace('.', '_')
+
+    def _calculate_file_abi_tag_heuristic_windows(self, filename: str) -> Optional[mesonpy._tags.Tag]:
+        """Try to calculate the Windows tag from the Python extension file name."""
+        match = _WINDOWS_NATIVE_MODULE_REGEX.match(filename)
+        if not match:
+            return None
+        tag = match.group('tag')
+
+        try:
+            return mesonpy._tags.StableABITag(tag)
+        except ValueError:
+            return mesonpy._tags.InterpreterTag(tag)
+
+    def _calculate_file_abi_tag_heuristic_posix(self, filename: str) -> Optional[mesonpy._tags.Tag]:
+        """Try to calculate the Posix tag from the Python extension file name."""
+        # sysconfig is not guaranted to export SHLIB_SUFFIX but let's be
+        # preventive and check its value to make sure it matches our expectations
+        try:
+            extension = sysconfig.get_config_vars().get('SHLIB_SUFFIX', '.so')
+            if extension != '.so':
+                raise NotImplementedError(
+                    f"We don't currently support the {extension} extension. "
+                    'Please report this to https://github.com/FFY00/mesonpy/issues '
+                    'and include information about your operating system.'
+                )
+        except KeyError:
+            warnings.warn(
+                'sysconfig does not export SHLIB_SUFFIX, so we are unable to '
+                'perform the sanity check regarding the extension suffix. '
+                'Please report this to https://github.com/FFY00/mesonpy/issues '
+                'and include the output of `python -m sysconfig`.'
+            )
+        match = _LINUX_NATIVE_MODULE_REGEX.match(filename)
+        if not match:  # this file does not appear to be a native module
+            return None
+        tag = match.group('tag')
+
+        try:
+            return mesonpy._tags.StableABITag(tag)
+        except ValueError:
+            return mesonpy._tags.InterpreterTag(tag)
+
+    def _calculate_file_abi_tag_heuristic(self, filename: str) -> Optional[mesonpy._tags.Tag]:
+        """Try to calculate the ABI tag from the Python extension file name."""
+        if os.name == 'nt':
+            return self._calculate_file_abi_tag_heuristic_windows(filename)
+        # everything else *should* follow the POSIX way, at least to my knowledge
+        return self._calculate_file_abi_tag_heuristic_posix(filename)
+
+    def _file_list_repr(self, files: Collection[str], prefix: str = '\t\t', max_count: int = 3) -> str:
+        if len(files) > max_count:
+            files = list(itertools.islice(files, max_count)) + [f'(... +{len(files)}))']
+        return ''.join(f'{prefix}- {file}\n' for file in files)
+
+    def _files_by_tag(self) -> Mapping[mesonpy._tags.Tag, Collection[str]]:
+        """Map files into ABI tags."""
+        files_by_tag: Dict[mesonpy._tags.Tag, List[str]] = collections.defaultdict(list)
+
+        for _, file in self._wheel_files['platlib']:
+            # if in platlib, calculate the ABI tag
+            tag = self._calculate_file_abi_tag_heuristic(file)
+            if tag:
+                files_by_tag[tag].append(file)
+
+        return files_by_tag
+
+    def _select_abi_tag(self) -> Optional[mesonpy._tags.Tag]:  # noqa: C901
+        """Given a list of ABI tags, selects the most specific one.
+
+        Raises an error if there are incompatible tags.
+        """
+        # Possibilities:
+        #   - interpreter specific (cpython/pypy/etc, version)
+        #   - stable abi (abiX)
+        tags = self._files_by_tag()
+        selected_tag = None
+        for tag, files in tags.items():
+            # no selected tag yet, let's assign this one
+            if not selected_tag:
+                selected_tag = tag
+            # interpreter tag
+            elif isinstance(tag, mesonpy._tags.InterpreterTag):
+                if tag != selected_tag:
+                    if isinstance(selected_tag, mesonpy._tags.InterpreterTag):
+                        raise ValueError(
+                            'Found files with incompatible ABI tags:\n'
+                            + self._file_list_repr(tags[selected_tag])
+                            + '\tand\n'
+                            + self._file_list_repr(files)
+                        )
+                    selected_tag = tag
+            # stable ABI
+            elif isinstance(tag, mesonpy._tags.StableABITag):
+                if isinstance(selected_tag, mesonpy._tags.StableABITag) and tag != selected_tag:
+                    raise ValueError(
+                        'Found files with incompatible ABI tags:\n'
+                        + self._file_list_repr(tags[selected_tag])
+                        + '\tand\n'
+                        + self._file_list_repr(files)
+                    )
+        return selected_tag
+
+    def _is_native(self, file: Union[str, pathlib.Path]) -> bool:
+        """Check if file is a native file."""
+        self._project.build()  # the project needs to be built for this :/
+
         with open(file, 'rb') as f:
-            return f.read(4) == b'\x7fELF'
+            if platform.system() == 'Linux':
+                return f.read(4) == b'\x7fELF'  # ELF
+            elif platform.system() == 'Darwin':
+                return f.read(4) in (
+                    b'\xfe\xed\xfa\xce',  # 32-bit
+                    b'\xfe\xed\xfa\xcf',  # 64-bit
+                    b'\xcf\xfa\xed\xfe',  # arm64
+                    b'\xca\xfe\xba\xbe',  # universal / fat (same as java class so beware!)
+                )
+            elif platform.system() == 'Windows':
+                return f.read(2) == b'MZ'
+
+        # For unknown platforms, check for file extensions.
+        _, ext = os.path.splitext(file)
+        if ext in ('.so', '.a', '.out', '.exe', '.dll', '.dylib', '.pyd'):
+            return True
+        return False
 
     def _warn_unsure_platlib(self, origin: pathlib.Path, destination: pathlib.Path) -> None:
         """Warn if we are unsure if the file should be mapped to purelib or platlib.
@@ -218,7 +430,7 @@ class _WheelBuilder():
         """
         # {moduledir_shared} is currently handled in heuristics due to a Meson bug,
         # but we know that files that go there are supposed to go to platlib.
-        if self._is_elf(origin) or destination.root == '{moduledir_shared}':
+        if self._is_native(origin):
             # The file is architecture dependent and does not belong in puredir,
             # so the warning is skipped.
             return
@@ -244,14 +456,13 @@ class _WheelBuilder():
                 if search_path.name == 'dist-packages' and search_path.parent.parent.name == 'lib':
                     calculated_path = origin.relative_to(search_path)
                     warnings.warn(f'File matched Debian heuristic ({calculated_path}): {origin} ({destination})')
-                    if not destination.root == '{moduledir_shared}':
-                        self._warn_unsure_platlib(origin, destination)
+                    self._warn_unsure_platlib(origin, destination)
                     return 'platlib', calculated_path
         # Try to map to the interpreter purelib or platlib
         for scheme in ('purelib', 'platlib'):
             # try to match the install path on the system to one of the known schemes
             scheme_path = pathlib.Path(sys_paths[scheme]).absolute()
-            destdir_scheme_path = self._project._install_dir / scheme_path.relative_to(scheme_path.anchor)
+            destdir_scheme_path = self._install_dir / scheme_path.relative_to(scheme_path.anchor)
             try:
                 wheel_path = pathlib.Path(origin).relative_to(destdir_scheme_path)
             except ValueError:
@@ -310,15 +521,15 @@ class _WheelBuilder():
 
         return wheel_files
 
-    def _install_file(
+    def _install_path(
         self,
         wheel_file: wheel.wheelfile.WheelFile,  # type: ignore[name-defined]
         counter: mesonpy._util.CLICounter,
         origin: Path,
         destination: pathlib.Path,
     ) -> None:
-        """"Install" file into the wheel and do the necessary processing before
-        doing so.
+        """"Install" file or directory into the wheel
+        and do the necessary processing before doing so.
 
         Some files might need to be fixed up to set the RPATH to the internal
         library directory on Linux wheels for eg.
@@ -327,34 +538,39 @@ class _WheelBuilder():
         counter.update(location)
 
         # fix file
-        if platform.system() == 'Linux':
-            # add .mesonpy.libs to the RPATH of ELF files
-            if self._is_elf(os.fspath(origin)):
-                # copy ELF to our working directory to avoid Meson having to regenerate the file
-                new_origin = self._libs_build_dir / pathlib.Path(origin).relative_to(self._project._build_dir)
-                os.makedirs(new_origin.parent, exist_ok=True)
-                shutil.copy2(origin, new_origin)
-                origin = new_origin
-                # add our in-wheel libs folder to the RPATH
-                elf = mesonpy._elf.ELF(origin)
-                libdir_path = f'$ORIGIN/{os.path.relpath(f".{self._project.name}.mesonpy.libs", destination.parent)}'
-                if libdir_path not in elf.rpath:
-                    elf.rpath = [*elf.rpath, libdir_path]
+        if os.path.isdir(origin):
+            for root, dirnames, filenames in os.walk(str(origin)):
+                # Sort the directory names so that `os.walk` will walk them in a
+                # defined order on the next iteration.
+                dirnames.sort()
+                for name in sorted(filenames):
+                    path = os.path.normpath(os.path.join(root, name))
+                    if os.path.isfile(path):
+                        arcname = os.path.join(destination, os.path.relpath(path, origin).replace(os.path.sep, '/'))
+                        wheel_file.write(path, arcname)
+        else:
+            if self._has_internal_libs and platform.system() == 'Linux':
+                # add .mesonpy.libs to the RPATH of ELF files
+                if self._is_native(os.fspath(origin)):
+                    # copy ELF to our working directory to avoid Meson having to regenerate the file
+                    new_origin = self._libs_build_dir / pathlib.Path(origin).relative_to(self._build_dir)
+                    os.makedirs(new_origin.parent, exist_ok=True)
+                    shutil.copy2(origin, new_origin)
+                    origin = new_origin
+                    # add our in-wheel libs folder to the RPATH
+                    elf = mesonpy._elf.ELF(origin)
+                    libdir_path = f'$ORIGIN/{os.path.relpath(f".{self._project.name}.mesonpy.libs", destination.parent)}'
+                    if libdir_path not in elf.rpath:
+                        elf.rpath = [*elf.rpath, libdir_path]
 
-        wheel_file.write(origin, location)
+            wheel_file.write(origin, location)
 
-    def build(
-        self,
-        sources: Dict[str, Dict[str, Any]],
-        copy_files: Dict[str, str],
-        directory: Path,
-    ) -> pathlib.Path:
+    def build(self, directory: Path) -> pathlib.Path:
         import wheel.wheelfile
 
         self._project.build()  # ensure project is built
 
         wheel_file = pathlib.Path(directory, f'{self.name}.whl')
-        wheel_files = self._map_to_wheel(sources, copy_files)
 
         with wheel.wheelfile.WheelFile(wheel_file, 'w') as whl:
             # add metadata
@@ -362,36 +578,34 @@ class _WheelBuilder():
             whl.writestr(f'{self.distinfo_dir}/WHEEL', self.wheel)
 
             # add license (see https://github.com/FFY00/meson-python/issues/88)
-            if self._project._metadata:
-                license_ = self._project._metadata.license
-                if license_ and license_.file:
-                    whl.write(
-                        self._project._source_dir / license_.file,
-                        f'{self.distinfo_dir}/{os.path.basename(license_.file)}',
-                    )
+            if self._project.license_file:
+                whl.write(
+                    self._source_dir / self._project.license_file,
+                    f'{self.distinfo_dir}/{os.path.basename(self._project.license_file)}',
+                )
 
             print('{light_blue}{bold}Copying files to wheel...{reset}'.format(**_STYLES))
             with mesonpy._util.cli_counter(
-                len(list(itertools.chain.from_iterable(wheel_files.values()))),
+                len(list(itertools.chain.from_iterable(self._wheel_files.values()))),
             ) as counter:
                 # install root scheme files
-                root_scheme = 'purelib' if self._project.is_pure else 'platlib'
-                for destination, origin in wheel_files[root_scheme]:
-                    self._install_file(whl, counter, origin, destination)
+                root_scheme = 'purelib' if self.is_pure else 'platlib'
+                for destination, origin in self._wheel_files[root_scheme]:
+                    self._install_path(whl, counter, origin, destination)
 
                 # install bundled libraries
-                for destination, origin in wheel_files['mesonpy-libs']:
+                for destination, origin in self._wheel_files['mesonpy-libs']:
                     assert platform.system() == 'Linux', 'Bundling libraries in wheel is currently only supported in POSIX!'
                     destination = pathlib.Path(f'.{self._project.name}.mesonpy.libs', destination)
-                    self._install_file(whl, counter, origin, destination)
+                    self._install_path(whl, counter, origin, destination)
 
                 # install the other schemes
                 for scheme in self._SCHEME_MAP:
                     if scheme in (root_scheme, 'mesonpy-libs'):
                         continue
-                    for destination, origin in wheel_files[scheme]:
+                    for destination, origin in self._wheel_files[scheme]:
                         destination = pathlib.Path(self.data_dir, scheme, destination)
-                        self._install_file(whl, counter, origin, destination)
+                        self._install_path(whl, counter, origin, destination)
 
         return wheel_file
 
@@ -443,7 +657,7 @@ class Project():
         # write the native file
         native_file_data = textwrap.dedent(f'''
             [binaries]
-            python3 = '{sys.executable}'
+            python = '{sys.executable}'
         ''')
         native_file_mismatch = (
             not self._meson_native_file.exists()
@@ -529,6 +743,17 @@ class Project():
                     f'expected `{self._metadata.requires_python}`'
                 )
 
+    @cached_property
+    def _wheel_builder(self) -> _WheelBuilder:
+        return _WheelBuilder(
+            self,
+            self._source_dir,
+            self._install_dir,
+            self._build_dir,
+            self._install_plan,
+            self._copy_files,
+        )
+
     @functools.lru_cache(maxsize=None)
     def build(self) -> None:
         """Trigger the Meson build."""
@@ -612,7 +837,7 @@ class Project():
         assert isinstance(version, str)
         return version
 
-    @property  # type: ignore[misc]
+    @property
     @functools.lru_cache(maxsize=1)
     def metadata(self) -> bytes:  # noqa: C901
         """Project metadata."""
@@ -631,210 +856,24 @@ class Project():
         core_metadata.headers['Version'] = [self.version]
         return bytes(core_metadata)
 
+    @property
+    def license_file(self) -> Optional[pathlib.Path]:
+        if self._metadata:
+            license_ = self._metadata.license
+            if license_ and license_.file:
+                return pathlib.Path(license_.file)
+        return None
+
     @property
     def is_pure(self) -> bool:
         """Is the wheel "pure" (architecture independent)?"""
-        # XXX: I imagine some users might want to force the package to be
-        # non-pure, but I think it's better that we evaluate use-cases as they
-        # arise and make sure allowing the user to override this is indeed the
-        # best option for the use-case.
-        not_pure = ('{bindir}', '{libdir_shared}', '{libdir_static}', '{py_platlib}', '{moduledir_shared}')
-        for data_type, files in self._install_plan.items():
-            for entry in files.values():
-                if entry['destination'] is None:  # pragma: no cover
-                    continue
-                if any(key in entry['destination'] for key in not_pure):
-                    return False
-        return True
+        return bool(self._wheel_builder.is_pure)
 
     @property
     def pep621(self) -> bool:
         """Does the project use PEP 621 metadata?"""
         return self._pep621
 
-    @property
-    def python_tag(self) -> str:
-        selected_tag = self._select_abi_tag()
-        if selected_tag and selected_tag.python:
-            return selected_tag.python
-        return 'py3'
-
-    @property
-    def abi_tag(self) -> str:
-        selected_tag = self._select_abi_tag()
-        if selected_tag:
-            return selected_tag.abi
-        return 'none'
-
-    @property
-    def platform_tag(self) -> str:
-        if self.is_pure:
-            return 'any'
-        # XXX: Choose the sysconfig platform here and let something like auditwheel
-        #      fix it later if there are system dependencies (eg. replace it with a manylinux tag)
-        platform_ = sysconfig.get_platform()
-        parts = platform_.split('-')
-        if parts[0] == 'macosx' and parts[1] in ('11', '12'):
-            # Workaround for bug where pypa/packaging does not consider macOS
-            # tags without minor versions valid. Some Python flavors (Homebrew
-            # for example) on macOS started to do this in version 11, and
-            # pypa/packaging should handle things correctly from version 13 and
-            # forward, so we will add a 0 minor version to MacOS 11 and 12.
-            # https://github.com/FFY00/meson-python/issues/91
-            # https://github.com/pypa/packaging/issues/578
-            parts[1] += '.0'
-            platform_ = '-'.join(parts)
-        return platform_.replace('-', '_').replace('.', '_')
-
-    def _calculate_file_abi_tag_heuristic_windows(self, filename: str) -> Optional[mesonpy._tags.Tag]:
-        """Try to calculate the Windows tag from the Python extension file name."""
-        match = _WINDOWS_NATIVE_MODULE_REGEX.match(filename)
-        if not match:
-            return None
-        tag = match.group('tag')
-
-        try:
-            return mesonpy._tags.StableABITag(tag)
-        except ValueError:
-            return mesonpy._tags.WindowsInterpreterTag(tag)
-
-    def _calculate_file_abi_tag_heuristic_posix(self, filename: str) -> Optional[mesonpy._tags.Tag]:
-        """Try to calculate the Posix tag from the Python extension file name."""
-        # sysconfig is not guaranted to export SHLIB_SUFFIX but let's be
-        # preventive and check its value to make sure it matches our expectations
-        try:
-            extension = sysconfig.get_config_vars().get('SHLIB_SUFFIX', '.so')
-            if extension != '.so':
-                raise NotImplementedError(
-                    f"We don't currently support the {extension} extension. "
-                    'Please report this to https://github.com/FFY00/mesonpy/issues '
-                    'and include information about your operating system.'
-                )
-        except KeyError:
-            warnings.warn(
-                'sysconfig does not export SHLIB_SUFFIX, so we are unable to '
-                'perform the sanity check regarding the extension suffix. '
-                'Please report this to https://github.com/FFY00/mesonpy/issues '
-                'and include the output of `python -m sysconfig`.'
-            )
-        match = _LINUX_NATIVE_MODULE_REGEX.match(filename)
-        if not match:  # this file does not appear to be a native module
-            return None
-        tag = match.group('tag')
-
-        try:
-            return mesonpy._tags.StableABITag(tag)
-        except ValueError:
-            return mesonpy._tags.LinuxInterpreterTag(tag)
-
-    def _calculate_file_abi_tag_heuristic(self, filename: str) -> Optional[mesonpy._tags.Tag]:
-        """Try to calculate the ABI tag from the Python extension file name."""
-        if os.name == 'nt':
-            return self._calculate_file_abi_tag_heuristic_windows(filename)
-        # everything else *should* follow the POSIX way, at least to my knowledge
-        return self._calculate_file_abi_tag_heuristic_posix(filename)
-
-    def _file_list_repr(self, files: Collection[str], prefix: str = '\t\t', max_count: int = 3) -> str:
-        if len(files) > max_count:
-            files = list(itertools.islice(files, max_count)) + [f'(... +{len(files)}))']
-        return ''.join(f'{prefix}- {file}\n' for file in files)
-
-    def _files_by_tag(self) -> Mapping[mesonpy._tags.Tag, Collection[str]]:
-        """Map files into ABI tags."""
-        files_by_tag: Dict[mesonpy._tags.Tag, List[str]] = collections.defaultdict(list)
-        for file, details in self._install_plan.get('targets', {}).items():
-            destination = pathlib.Path(details['destination'])
-            from_heuristic = False
-
-            # if in platlib, calculate the ABI tag
-            if not (
-                mesonpy._compat.is_relative_to(destination, '{py_platlib}')
-                or mesonpy._compat.is_relative_to(destination, '{moduledir_shared}')
-            ):
-                # XXX: Ideally we would just check the anchor placeholder in the
-                #      path (eg. check if it is {py_platlib}) but Meson seems to
-                #      have a bug where it sometimes returns a full path without
-                #      placeholder, so we will check if the file path is
-                #      relative to platlib.
-                #      This can be problematic because the platlib path might be
-                #      same one used for other schemes. In these situations we
-                #      will be picking up files are not supposed to be in
-                #      platlib, and that could just be supporting files.
-                #      See https://github.com/FFY00/meson-python/issues/95
-                #      Meson bug: https://github.com/mesonbuild/meson/issues/10601
-                sys_vars = sysconfig.get_config_vars().copy()
-                sys_vars['base'] = sys_vars['platbase'] = sys.base_prefix
-                platlib = sysconfig.get_path('platlib', vars=sys_vars)
-                if platlib and mesonpy._compat.is_relative_to(destination, platlib):
-                    from_heuristic = True
-                else:
-                    continue
-
-            tag = self._calculate_file_abi_tag_heuristic(file)
-            if tag:
-                if from_heuristic:
-                    warnings.warn(
-                        'Could not tell with certainty if this file was meant '
-                        'to be mapped to platlib, but it was used to calculate '
-                        f'the ABI tag: {destination}'
-                    )
-                files_by_tag[tag].append(file)
-
-        return files_by_tag
-
-    def _select_abi_tag(self) -> Optional[mesonpy._tags.Tag]:  # noqa: C901
-        """Given a list of ABI tags, selects the most specific one.
-
-        Raises an error if there are incompatible tags.
-        """
-        # Possibilities:
-        #   - interpreter specific (cpython/pypy/etc, version)
-        #   - stable abi (abiX)
-        tags = self._files_by_tag()
-        selected_tag = None
-        for tag, files in tags.items():
-            if __debug__:  # sanity check
-                if os.name == 'nt':
-                    assert not isinstance(tag, mesonpy._tags.LinuxInterpreterTag)
-                else:
-                    assert not isinstance(tag, mesonpy._tags.WindowsInterpreterTag)
-            # no selected tag yet, let's assign this one
-            if not selected_tag:
-                selected_tag = tag
-            # interpreter tags
-            elif isinstance(tag, mesonpy._tags.LinuxInterpreterTag):
-                if tag != selected_tag:
-                    if isinstance(selected_tag, mesonpy._tags.LinuxInterpreterTag):
-                        raise ValueError(
-                            'Found files with incompatible ABI tags:\n'
-                            + self._file_list_repr(tags[selected_tag])
-                            + '\tand\n'
-                            + self._file_list_repr(files)
-                        )
-                    selected_tag = tag
-            elif isinstance(tag, mesonpy._tags.WindowsInterpreterTag):
-                if tag != selected_tag:
-                    if isinstance(selected_tag, mesonpy._tags.WindowsInterpreterTag):
-                        warnings.warn(
-                            'Found files with different ABI tags but couldn\'t tell '
-                            'if they are incompatible:\n'
-                            + self._file_list_repr(tags[selected_tag])
-                            + '\tand\n'
-                            + self._file_list_repr(files)
-                            + 'Please report this to https://github.com/FFY00/mesonpy/issues.'
-                        )
-                    selected_tag = tag
-            # stable ABI
-            elif isinstance(tag, mesonpy._tags.StableABITag):
-                if isinstance(selected_tag, mesonpy._tags.StableABITag) and tag != selected_tag:
-                    raise ValueError(
-                        'Found files with incompatible ABI tags:\n'
-                        + self._file_list_repr(tags[selected_tag])
-                        + '\tand\n'
-                        + self._file_list_repr(files)
-                    )
-        return selected_tag
-
     def sdist(self, directory: Path) -> pathlib.Path:
         """Generates a sdist (source distribution) in the specified directory."""
         # generate meson dist file
@@ -863,7 +902,9 @@ class Project():
                     continue
 
                 info = tarfile.TarInfo(member.name)
-                info.size = os.path.getsize(path)
+                file_stat = os.stat(path)
+                info.size = file_stat.st_size
+                info.mode = int(oct(file_stat.st_mode)[-3:], 8)
 
                 # rewrite the path if necessary, to match the sdist distribution name
                 if dist_name != meson_dist_name:
@@ -886,7 +927,7 @@ class Project():
 
     def wheel(self, directory: Path) -> pathlib.Path:  # noqa: F811
         """Generates a wheel (binary distribution) in the specified directory."""
-        wheel = _WheelBuilder(self).build(self._install_plan, self._copy_files, self._build_dir)
+        wheel = self._wheel_builder.build(self._build_dir)
 
         final_wheel = pathlib.Path(directory, wheel.name)
         shutil.move(os.fspath(wheel), final_wheel)
